@@ -1,33 +1,34 @@
+from os import getcwd
+from os.path import join
+from textwrap import dedent
+from time import time
+from traceback import format_exc
+
+from pandas import read_csv, DateOffset
+from PyQt5.QtTest import QTest
+from PyQt5.QtWidgets import QApplication
+
 from kiwoom import config
 from kiwoom.config import history
-from kiwoom.config.types import ExitType
 from kiwoom.config.screen import Screen
+from kiwoom.config.types import ExitType
 from kiwoom.core.kiwoom import Kiwoom
 from kiwoom.core.server import Server
 from kiwoom.data.share import Share
 from kiwoom.utils.general import *
-from kiwoom.utils.manager import timer, Downloader
-
-from PyQt5.QtWidgets import QApplication
-from PyQt5.QtTest import QTest
-from time import time
-from os import getcwd
-from os.path import join
-from textwrap import dedent
-from traceback import format_exc
-from pandas import read_csv, DateOffset, Timestamp
+from kiwoom.utils.manager import Downloader, timer
 
 
 class Bot:
     def __init__(self, server=None):
-        self.api = Kiwoom()
-        self.scr = Screen()
-        self.share = Share()
+        self.api: Kiwoom = Kiwoom()
+        self.scr: Screen = Screen()
+        self.share: Share = Share()
 
         # Connect server as a slot
-        self.server = server if issubclass(type(server), Server) else Server()
+        self.server: Server = server if issubclass(type(server), Server) else Server()
+        self.server.init(bot=self, api=self.api, share=self.share)
         self.default_connect(server)
-        self.server.init(api=self.api, share=self.share)
 
     def default_connect(self, server):
         self.api.set_connect_hook('on_receive_tr_data', 'rq_name')
@@ -51,7 +52,7 @@ class Bot:
             return True
         return False
 
-    def stock_list(self, market):
+    def stock_list(self, market) -> list[str]:
         """
         Returns all stock codes in the given market
 
@@ -59,10 +60,13 @@ class Bot:
             one of market code in kiwoom.config.markets
         """
         market = str(market)
-        if history.is_market(market):
-            return sorted(self.api.get_code_list_by_market(market).split(';')[:-1])
+        codes = self.api.get_code_list_by_market(market)
+        codes = codes.split(';')[:-1]
+        if not codes:
+            raise ValueError(f'No stock codes for {market=}.')
+        return sorted(codes)
 
-    def sector_list(self, market_gubun):
+    def sector_list(self, market_gubun) -> list[str]:
         """
         Returns all sector codes in the given market_gubun
 
@@ -127,6 +131,7 @@ class Bot:
 
             # To share variables with Slot
             kwargs = effective_args(locals(), remove=['ctype', 'tr_code'])
+            self.share.remove_single(name())
             self.share.update_single(name(), 'error', False)
             self.share.update_single(name(), 'restart', False)
             self.share.update_single(name(), 'complete', False)
@@ -177,7 +182,7 @@ class Bot:
                                     self.share.update_single(name(), 'complete', True)
                                     return
 
-                    else:  # if period in ['day', 'week', 'year']
+                    else:  # if period in ['day', 'week', 'month', 'year']
                         last_day = date(df.index[-1])
                         # To push 'start' date further as much as possible. If None, set newly.
                         if 'start' not in kwargs or date(kwargs['start']) <= last_day:
@@ -190,8 +195,8 @@ class Bot:
                                 self.share.update_single(name(), 'complete', True)
                                 return
 
-                    # Resolve memory issues
-                    del df
+                    # Once read, use later in Server.history_to_csv() to increase efficiency
+                    self.share.update_single(name(), 'file', df)
 
                 # If any exception, just skip
                 except Exception as err:
@@ -211,15 +216,16 @@ class Bot:
             Start downloading.
         """
         # Check requesting status
-        self.share.single['histories']['nrq'] += 1
-        if history.SPEEDING:
-            if self.share.get_single('histories', 'nrq') >= history.REQUEST_LIMIT_TRY:
-                # Set back to default configuration
-                if self.share.get_single('histories', 'cnt') == 0:
-                    self.share.update_single('history', 'impossible', True)
-                self.share.update_single(name(), 'restart', True)
-                self.api.unloop()
-                return
+        if 'histories' in self.share.single:
+            self.share.single['histories']['nrq'] += 1
+            if history.SPEEDING:
+                if self.share.get_single('histories', 'nrq') >= history.REQUEST_LIMIT_TRY:
+                    # Set back to default configuration
+                    if self.share.get_single('histories', 'cnt') == 0:
+                        self.share.update_single(name(), 'impossible', True)
+                    self.share.update_single(name(), 'restart', True)
+                    self.api.unloop()
+                    return
 
         # Finally request data to server
         for key, val in history.inputs(tr_code, code, unit, end):
@@ -228,7 +234,8 @@ class Bot:
 
         # If comm_rq_data returns non-zero error code, restart downloading
         if self.api.comm_rq_data(name(), tr_code, prev_next, scr_no) != 0:
-            self.share.update_single('history', 'restart', True)
+            self.share.update_single(name(), 'impossible', True)
+            self.share.update_single(name(), 'restart', True)
             self.api.unloop()
             return
 
@@ -286,8 +293,9 @@ class Bot:
             path = getcwd()
 
         # Initialize status
-        self.share.update_single(name(), 'nrq', 0)  # number of request
-        self.share.update_single(name(), 'cnt', 0)  # number of stocks
+        self.share.remove_single(name())
+        self.share.single[name()]['nrq'] = 0  # number of request
+        self.share.single[name()]['cnt'] = 0  # number of stocks
         self.share.update_args(name(), effective_args(locals()))
 
         """
@@ -298,12 +306,24 @@ class Bot:
         if not any([market, sector]) or all([market, sector]):
             raise RuntimeError("Download target must be either of 'market' or 'sector'.")
         elif market is not None:
-            lst, ctype, mname = self.stock_list(market), str(history.STOCK).lower(), history.MARKETS[market]
+            market = market.upper()
+            ctype, mname = str(history.STOCK).lower(), history.MARKETS[market]
+            if market != 'NXT':
+                krx = set(self.stock_list(market))
+                nxt = set(self.stock_list('NXT'))
+                nxt.intersection_update(krx)
+                sor = set([c + '_AL' for c in list(nxt)])
+                lst = (krx - nxt) | sor
+                lst = sorted(list(lst))
+            else:
+                lst = self.stock_list(market)
+                lst = [c + '_AL' for c in lst]
         elif sector is not None:
-            lst, ctype, mname = self.sector_list(sector), str(history.SECTOR).lower(), history.MARKET_GUBUNS[sector]
-
+            ctype, mname = str(history.SECTOR).lower(), history.MARKET_GUBUNS[sector]
+            lst = self.sector_list(sector)
+        
         # Set the portion in download list
-        from_, to_ = None, None
+        from_, to_ = 0, None
         if all([slice, code]):
             raise RuntimeError("Only one option is available: either of 'slice' or 'start_code'.")
         # Option1 - Slice
@@ -315,11 +335,9 @@ class Bot:
         # Option2 - Code
         elif code is not None:
             from_, to_ = lst.index(code), None
-        # Fully download
-        else:
-            from_, to_ = 0, None
 
         # Select target in download list
+        tot = len(lst)
         lst = lst[from_: to_]
 
         # To print progress bar
@@ -339,8 +357,8 @@ class Bot:
 
         for i, code in enumerate(lst):
             if i % divisor == 0:
-                pct = (i / len(lst)) * 100
-                print(f'\nDownloading ..\t{pct: .1f}% ({i} of {len(lst)})')
+                pct = ((from_ + i) / tot) * 100
+                print(f'Downloading ..\t{pct: .1f}% ({from_ + i} of {tot})')
 
             # Try downloading
             try:
@@ -384,7 +402,7 @@ class Bot:
             # 6) Successfully downloaded but exceeds request limit items
             if history.SPEEDING:
                 if self.share.single[name()]['cnt'] >= history.REQUEST_LIMIT_ITEM:
-                    # To check whether items are downloaded by the actual requests
+                    # To check whether items are downloaded by the actual requests to download more if possible
                     if self.share.single[name()]['nrq'] >= history.REQUEST_LIMIT_ITEM:
                         break
 
@@ -392,9 +410,10 @@ class Bot:
             Close downloading
         """
         cnt = self.share.get_single(name(), 'cnt')
+        cum = from_ + cnt
         msg = dedent(
             f"""
-            Download Done for {100 * cnt / len(lst) if lst else 100: .1f}% ({cnt} of {len(lst)}) {ctype}s in {mname}.
+            Download Done for {100 * cum / tot if lst else 100: .1f}% ({cum} of {tot}) {ctype}s in {mname}.
             Download Time : {(time() - begin) / 60: .1f} minutes (with {self.share.single[name()]['nrq']} requests)\n
             """
         ) + status
@@ -404,7 +423,7 @@ class Bot:
         if cnt == len(lst):
             return ExitType.SUCCESS
         # Else return remaining items
-        return from_ + cnt, to_
+        return cum, to_
 
     def exit(self, ecode=0):
         """
