@@ -1,26 +1,31 @@
-from kiwoom import config
-from kiwoom.config import history
-from kiwoom.config.error import msg
-from kiwoom.config.types import MULTI
-from kiwoom.data.preps import string
-from kiwoom.utils.general import name, date
-from kiwoom.utils.manager import Downloader
-
 from os import getcwd, makedirs
-from os.path import join, exists
+from os.path import exists, join
 from textwrap import dedent
 from traceback import format_exc
 from warnings import warn
 
 import pandas as pd
 
+from kiwoom import config
+from kiwoom.config import history
+from kiwoom.config.error import msg
+from kiwoom.config.types import MULTI
+from kiwoom.core.kiwoom import Kiwoom
+from kiwoom.data.share import Share
+from kiwoom.data.preps import string
+from kiwoom.utils.general import date, name
+from kiwoom.utils.manager import Downloader
+
 
 class Server:
     def __init__(self):
-        self.api = None
-        self.share = None
+        from kiwoom.core.bot import Bot
+        self.bot: Bot = None
+        self.api: Kiwoom = None
+        self.share: Share = None
 
-    def init(self, api, share):
+    def init(self, bot, api, share):
+        self.bot = bot
         self.api = api
         self.share = share
 
@@ -34,10 +39,10 @@ class Server:
         """
         Default slot for 'on_event_connect'
 
-        When Kiwoom.on_event_connect(...) is called, this method automatically will be called.
+        When Kiwoom.on_event_connect(...) is called, this method will be automatically called.
         """
         print(f'\n로그인 {msg(err_code)}')
-        print(f'\n* 시스템 점검\n  - 월 ~ 토 : 05:05 ~ 05:10\n  - 일 : 04:00 ~ 04:30\n')
+        print(f'\n* 시스템 점검\n  - 월 ~ 토 : 06:50 ~ 06:55\n  - 일 : 05:00 ~ 05:15\n')
         self.api.unloop()
 
     # Default event slot for on_receive_msg_slot
@@ -116,43 +121,44 @@ class Server:
                 Make time-related column as pandas Datetime index
             """
             # To handle exceptional time and dates
-            if not df.empty and history.is_sector(code) and col == '체결시간':
+            if not df.empty \
+               and history.is_sector(code) \
+               and col == '체결시간':
+
                 # To choose exceptional datetime replacer
-                edrfec = history.EXCEPTIONAL_DATETIME_REPLACER_FOR_EXCEPTIONAL_CODE
-                replacer = edrfec[code] if code in edrfec else history.EXCEPTIONAL_DATETIME_REPLACER
+                replacer = history.EXCEPTIONAL_DATETIME_REPLACER
+                if code in history.EXCEPTIONAL_DATETIME_REPLACER_FOR_EXCEPTIONAL_CODE:
+                    replacer = history.EXCEPTIONAL_DATETIME_REPLACER_FOR_EXCEPTIONAL_CODE[code]
 
-                # Find index of dates that delayed market opening time and inconvertibles in df
-                indices = dict()
-                exceptions = list()
-                start, end = date(df[col].iat[0][:len('YYYYMMDD')]), date(df[col].iat[-1][:len('YYYYMMDD')])
-                for ymd, delay in history.EXCEPTIONAL_DATES.items():
+                # To handle delayed market openings
+                delayed = dict()
+                start = date(df[col].iat[0][:len('YYYYMMDD')])
+                end = date(df[col].iat[-1][:len('YYYYMMDD')])
+                for ymd, hour in history.DELAYED_MARKET_OPENING.items():
                     if start <= date(ymd) <= end:
-                        day = df[col].loc[df[col].str.match(ymd)]
-                        indices[ymd] = day.index
-
-                        # To save original data
-                        for regex, datetime in replacer.items():
-                            series = day.loc[day.str.contains(regex, regex=True)]
-                            series = series.replace(regex={regex: datetime})
-                            series = pd.to_datetime(series, format='%Y%m%d%H%M%S')
-                            exceptions.append(series)
+                        delayed_replacer = dict(replacer)
+                        for regex, hhmmss in delayed_replacer.items():
+                            hhmmss = min(int(hhmmss) + (hour * 10000), 180000)
+                            hhmmss = str(hhmmss).zfill(6)
+                            delayed_replacer[regex] = hhmmss
+                        
+                        for regex, hhmmss in delayed_replacer.items():
+                            data = df.loc[df[col].str.match(ymd), col]
+                            target = data.loc[data.str.contains(regex, regex=True)]
+                            target = target.replace(regex={regex: hhmmss})
+                            for idx, tsp in target.items():
+                                delayed[idx] = pd.to_datetime(tsp, format=fmt)
 
                 # Replace inconvertibles (888888, 999999) to (16:00:00, 18:00:00)
-                df[col].replace(regex=replacer, inplace=True)
+                df[col] = df[col].replace(regex=replacer)
 
                 # To make column as pandas datetime series
                 df[col] = pd.to_datetime(df[col], format=fmt)
 
-                # Subtract delayed market time as if it pretends to start normally
-                for ymd, idx in indices.items():
-                    delay = history.EXCEPTIONAL_DATES[ymd]
-                    df.loc[idx, col] -= pd.DateOffset(hours=delay)
+                # Handle delayed market openings
+                df.loc[list(delayed.keys()), col] = list(delayed.values())
 
-                # Replace subtracted exceptional times back to original
-                for series in exceptions:
-                    df.loc[series.index, col] = series
-
-            # col='일자' or including df.empty for both col
+            # To handle etc.
             else:
                 df[col] = pd.to_datetime(df[col], format=fmt)
 
@@ -169,16 +175,14 @@ class Server:
             if 'end' in kwargs:
                 df = df.loc[:kwargs['end']]
 
-            # If server sent mixed data
-            if not df.index.is_monotonic_increasing:
-                raise RuntimeError(
-                    f'Downloaded data is not monotonic increasing. Error at Server.history() with code={code}.'
-                )
-
             # Rename column
             if period == 'tick':
                 df.rename(columns={'현재가': '체결가'}, inplace=True)
 
+            # Validate data
+            if not df.index.is_monotonic_increasing:
+                df = df.sort_index(kind='stable')
+            
             # Save data to csv file
             self.history_to_csv(df, code, kwargs['path'], kwargs['merge'], kwargs['warning'])
 
@@ -231,12 +235,18 @@ class Server:
                 col = df.index.name
                 if col not in ['일자', '체결시간']:
                     raise ValueError(f"No column matches '일자' or '체결시간'. Merge can't be done.")
-                db = pd.read_csv(
-                    file,
-                    index_col=[col],
-                    parse_dates=[col],
-                    encoding=config.ENCODING
-                )
+
+                # Existing data
+                if 'file' in self.share.single['history']:
+                    db = self.share.get_single('history', 'file')
+                else:
+                    # Read the existing file from disk
+                    db = pd.read_csv(
+                        file,
+                        index_col=[col],
+                        parse_dates=[col],
+                        encoding=config.ENCODING
+                    )
                 db.dropna(axis='index', inplace=True)
 
                 if not db.empty:
@@ -284,11 +294,14 @@ class Server:
                         # Just concatenate if no overlapping period.
                         df = pd.concat([db, df], axis=0, join='outer', copy=False)
 
-        if not df.index.is_monotonic_increasing:
-            raise RuntimeError(
-                f'Error at Server.history_to_csv(file={file}, ...)/\n'
-                + 'File to write is not monotonic increasing with respect to time.'
-            )
+        # == Notice : Does not raise error anymore, sort df by time instead!  ==
+        # if not df.index.is_monotonic_increasing:
+        #     # Write data to check flaws
+        #     df.to_csv(file + '.err', encoding=config.ENCODING)
+        #     raise RuntimeError(
+        #         f'Error at Server.history_to_csv(file={file}, ...)/\n'
+        #         + 'File to write is not monotonic increasing with respect to time.'
+        #     )
 
         # To prevent overwriting
         if not merge and exists(file):
