@@ -1,5 +1,4 @@
 from PyQt5.QAxContainer import QAxWidget
-from kiwoom.core.connector import Connector
 
 
 class API(QAxWidget):
@@ -349,7 +348,7 @@ class API(QAxWidget):
         9 : 최유리지정가(IOC)
         A : 최유리지정가(FOK)
         """
-        fn = "SendOrderFO(QString, QString, QString, Int, QString, Int, Int, QString, QString)"
+        fn = "SendOrderFO(QString, QString, QString, QString, Int, QString, QString, Int, QString, QString)"
         args = [rq_name, scr_no, acc_no, code, ord_kind, sl_by_tp, ord_tp, qty, price, org_ord_no]
         return self.call(fn, args)
 
